# Модули, функции и Mix

### Дефиниране на модул

```elixir
defmodule Times do
  def double(n) do
    n * 2
  end
end
```



### do..end блокове с малко код

```elixir
# Кратък синтаксис за one-liners
def double(n), do: n * 2
# do..end синтаксис, позволява много редове код
def again_double(n) do
  n * 2
end
```



### Дефиниране на функции и pattern matching

```elixir
defmodule Factorial do
  def of(0), do: 1
  def of(n), do: n * of(n - 1)
end
```



### Редът на дефиниране на "главите" на функцията има значение

```elixir
defmodule Factorial do
  def of(n), do: n * of(n - 1)
  def of(0), do: 1
end
```



### Guard клаузи при дефиниране на функции

```elixir
defmodule Fibonacci do
  def of(1), do: 1
  def of(2), do: 1
  def of(n) when is_number(n), do: of(n - 1) + of(n - 2)
end
```



### Параметри със стойности по подразбиране

- дефинират се с `\\`
- = се използва за съпоставяне (pattern matching), затова е нужен друг синтаксис

```elixir
defmodule Example do
  def func(p1, p2 \\ 2, p3 \\ 3, p4) do
    IO.inspect [p1, p2, p3, p4]
  end
end

Example.func("a", "b") # => ["a", 2, 3, "b"]
Example.func("a", "b", "c") # => ["a", "b", 3, "c"]
Example.func("a", "b", "c", "d") # => ["a", "b", "c", "d"]
```



```elixir
Enum.join(
  String.split(
    String.upcase(
      "name,faculty_number,grade"),
    ","),
  " "
)
"NAME FACULTY_NUMBER GRADE"
```



### Pipe оператора - `|>`

```elixir
"name,faculty_number,grade"
|> String.upcase
|> String.split(",")
|> Enum.join(" ")
"NAME FACULTY_NUMBER GRADE"
```



### Връзки между модули и влагане

```elixir
defmodule Outer do
  defmodule Inner do
    def inner_func do
      "hello world"
    end
  end

  def outer_func do
    "Greeting from the inner func: #{Outer.Inner.inner_func}"
  end
end

Outer.outer_func # => "Greeting from the inner func: hello world"
Outer.Inner.inner_func # => "hello world"
```



```elixir
defmodule Deep.Outer do
  def outer, do: :outer

  defmodule Inner do
    # Трябва Deep.Outer.outer/1 да го реферираме директно,
    # ще гръмне
    def inner_outer, do: Outer.outer()

    # Няма да се компилира
    def inner_outer2, do: outer()
  end
end
```



##### Между другото нищо не ни пречи директно да направим това:

```elixir
defmodule Deep.Outer.Inner do
  def inner, do: :inner
end
```



### Използване на Erlang модули

```elixir
:rand.uniform(100) # => 98
:rand.uniform(100) # => 41
```


```elixir
defmodule CsvUtils do
  import String

  def upcase_space_transform(csv_line) do
    csv_line
    |> upcase
    |> split(",")
    |> Enum.map(&strip/1)
    |> Enum.join(" ")
  end
end
```


```elixir
# Using only:
defmodule CsvUtils do
  import String, only: [upcase: 1, split: 2, strip: 1]

  def upcase_space_transform(csv_line) do
    csv_line
    |> upcase
    |> split(",")
    |> Enum.map(&strip/1)
    |> Enum.join(" ")
  end
end
```


```elixir
# Using except:
defmodule RationalNumbers do
  import Kernel, except: [abs: 1]

  def abs({n, d}), do: {Kernel.abs(n), Kernel.abs(d)}
end
```



```elixir
defmodule Foo do
  def hi, do: "Hi from Foo"
end

defmodule Bar do
  def hi, do: "Hi from Bar"
end

defmodule Baz do
  import Foo
  import Bar

  def say_hi, do: hi()
end
# ** (CompileError) iex:16: function hi/0 imported from both Bar and Foo, call is ambiguous
#     (elixir 1.12.3) src/elixir_dispatch.erl:128: :elixir_dispatch.expand_import/6
#     (elixir 1.12.3) src/elixir_dispatch.erl:91: :elixir_dispatch.dispatch_import/5
#     iex:16: (module)
```



```elixir
defmodule Outer.Inner do
  def inner_func do
    "hello world"
  end
end

defmodule Outer do
  alias Outer.Inner, as: Inner

  def outer_func do
    "Greeting from the inner func: #{Inner.inner_func}"
  end
end
```



```elixir
defmodule Outer do
  defmodule Foo do
    def hi, do: "Hi from Foo"
  end

  defmodule Bar do
    def hi, do: "Hi from Bar"
  end
end

alias Outer.{Foo, Bar}
Foo.hi()
# => "Hi from Foo"

Bar.hi()
# => "Hi from Bar"
```



```elixir
iex(1)> Integer.is_odd(3)
** (CompileError) iex:1: you must require Integer before invoking the macro Integer.is_odd/1
    (elixir) src/elixir_dispatch.erl:97: :elixir_dispatch.dispatch_require/6
iex(1)> require Integer
Integer
iex(2)> Integer.is_odd(3)
true
iex(3)>
```



```elixir
defmodule Greeter do
  @moduledoc """
  A module designed to greet strangers and friends alike!
  """

  @standard_greeting "Hello, stranger!"

  @doc "Greet someone!"
  def greet(nil) do
    IO.puts @standard_greeting
  end

  def greet(name) do
    IO.puts "Hello, #{name}!"
  end
end
```




#### Могат да имат няколко глави като нормалните функции

```elixir
add = fn
  a, b when is_integer(a) and is_integer(b) -> a + b
  a, b when is_list(a) and is_list(b) -> a ++ b
  a, b when is_binary(a) and is_binary(b) -> <<a::binary, b::binary>>
end
res = add.(<<0::size(16)>>, <<1::size(16)>>)
<<0, 0, 0, 1>>
iex(10)> Kernel.bit_size(res)
32
```



#### Пример

```elixir
iex(5)> fun = &is_integer/1
&:erlang.is_integer/1
iex(6)> fun.(0)
true
```



#### Oще пример

```elixir
iex(19)> succ = & &1 + 1
#Function<6.127694169/1 in :erl_eval.expr/5>
iex(20)> succ.(1)
2
```


#### Oще пример с оператор

```elixir
iex(13)> doubler = &{&1, &1}
#Function<6.127694169/1 in :erl_eval.expr/5>
iex(14)> doubler.(1)
{1, 1}
```